#! /bin/bash -e


# Copyright (c) 2007 Takahiro Shinozaki

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Synopsis
#   htkclustering.bash [-debug] [-config config] [-doneflg doneflg]
#
# Description:
#	tree based triphone state clustering.
#	this script takes an initial clustering threshold and desired 
#	number of states as the arguments.

# Change log:
# Tue Mar 26 21:05:39 JST 2013
#   A bug fix in threshold update
# Thu Dec 27 16:29:13 JST 2007
#   The initial version was written by Takahiro Shinozaki.

[ -n "$debug" ] && set -x
scriptp=$0
script=`basename $0`

function usage {
    echo "Usage: $script [-debug]"
    echo "          [-config CONFIGURATION]"
    echo "          [-doneflg DONE_FLAG]"
    exit 1
}

# initialize local variables
doneflg=done
envlog=env.log

while [ -n "$1" ]; do
    case "$1" in
	-debug) shift; export debug=1; set -x;;
	-config) shift; export config=$1; shift; . $config;;
	-doneflg) shift; doneflg="$1"; shift;;
	*) usage;
    esac
done

if [ -f $doneflg ]; then
    echo "Nothing to be done ($script)."
    exit 0
fi

# dump environmental variables
env > $envlog

# print process info
echo --htkclustering.bash START--
date
pwd
if [ ! -e $script ]; then
    cp -p $scriptp .
fi

# All the input global variables used in this script are (should be) listed here.
echo debug=$debug
echo config=$config
echo HHEd=$HHed
echo corpustri=$corpustri
echo alltri=$alltri
echo orghmm=$orghmm
echo stats=$stats
echo initthr=$initthr
echo nstates=$nstates
echo tdctemplate=$tdctemplate
echo brun=$brun
echo bruncfg=$bruncfg

# check requirments
if [ "$HHEd" = "" ]; then echo HHEd is not defined; exit 1; fi
if [ "$corpustri" = "" ]; then echo corpustri is not defined; exit 1; fi
if [ "$alltri" = "" ]; then echo alltri is not defined; exit 1; fi
if [ "$orghmm" = "" ]; then echo orghmm is not defined; exit 1; fi
if [ "$stats" = "" ]; then echo stats is not defined; exit 1; fi
if [ "$initthr" = "" ]; then echo initthr is not defined; exit 1; fi
if [ "$nstates" = "" ]; then echo nstates is not defined; exit 1; fi
if [ "$tdctemplate" = "" ]; then echo tdctemplate is not defined; exit 1; fi
if [ "$brun" = "" ]; then echo brun is not defined; exit 1; fi


if [ -n "$bruncfg" ]; then
    brunopt="-c $bruncfg"
else
    brunopt=""
fi

function clustering {
    thre=$1
    outd=$2

    mkdir -p $outd
    STATS_FILE=$stats
    THRESHOLD=$thre
    PHYSICALTRI=$alltri
    . $tdctemplate > $outd/cluster.hed

    # state clustering
    out=$outd/tdc.out
    err=$outd/tdc.err
    cmd="$HHEd -T 00407 -D -H $orghmm -M $outd $outd/cluster.hed $corpustri"
    echo $out $err $cmd | $brun $brunopt -

    if [ -e tdcTree ]; then
	mv tdcTree $outd
    else
	echo "Filename for tree must be tdcTree (if it is needed)."
    fi
    # count number of states (It is equal to num of means at this stage)
    nst=`grep MEAN $outd/hmmdefs | wc -l`
    echo $thre $nst
    # return nst
}

# try clustring with different thresholds until desired number of states is obtained.

thre=$initthr
outd=th$thre
clustering $thre $outd
if [ $(((($nst-$nstates)*100*2)/$nstates)) -eq 0 ]; then # permit +/-0.5%
    mv $outd st$nstates
    touch $doneflg
    exit 0
fi

nstp=$nst; threp=$thre
thre=$(($thre+$nst-$nstates))
outd=th$thre
clustering $thre $outd
if [ $(((($nst-$nstates)*100*2)/$nstates)) -eq 0 ]; then # permit +/-0.5%
    mv $outd st$nstates
    touch $doneflg
    exit 0
fi
                                                                                                             
for i in `seq 1 10`; do
    nstpp=$nstp; nstp=$nst; threpp=$threp; threp=$thre;
    thre=$(( (($threp-$threpp)*($nstates-$nstp))/($nstp-$nstpp) + $threp ))
    if [ $thre -eq $threp ]; then
	break
    fi
    outd=th$thre
    clustering $thre $outd
    if [ $(((($nst-$nstates)*100*2)/$nstates)) -eq 0 ]; then # permit +/-0.5%
        break
    fi
done
if [ $(((($nst-$nstates)*100*2)/$nstates)) -ne 0 ]; then # permit +/-0.5%
#    echo "WARN large error in number of states"
    echo "ERROR: large error in number of states"
    exit 1
fi
mv $outd st$nstates
touch $doneflg
echo --$script END--
